MODULE AnisotropicMaterialModels

USE DefUtils
IMPLICIT NONE

CONTAINS

!------------------------------------------------------------------------------
!> Return second invariant.
!> Note: Actually SQUARE of the second invariant of velocity is returned
!> Note: Copied from MaterialModels.F90 
!------------------------------------------------------------------------------
   FUNCTION SecondInvariant( Velo,dVelodx,CtrMetric,Symb ) RESULT(SecInv)
!------------------------------------------------------------------------------

     REAL(KIND=dp), OPTIONAL :: CtrMetric(3,3),Symb(3,3,3)
     REAL(KIND=dp) :: Velo(3),dVelodx(3,3),SecInv

!------------------------------------------------------------------------------

     INTEGER :: i,j,k,l
     REAL(KIND=dp) :: CovMetric(3,3),s,t

     SecInv = 0.0D0

     IF ( CurrentCoordinateSystem() == Cartesian ) THEN
!------------------------------------------------------------------------------

       DO i=1,3
         DO j=1,3
           s = dVelodx(i,j) + dVelodx(j,i)
           SecInv = SecInv + s * s
         END DO
       END DO
!------------------------------------------------------------------------------
     ELSE IF ( CurrentCoordinateSystem() == AxisSymmetric ) THEN

        SecInv = (2*dVelodx(1,1))**2 + (2*dVelodx(2,2))**2 + &
          2*(dVelodx(1,2) + dVelodx(2,1))**2 + (2*Velo(1)*symb(1,3,3))**2

     ELSE

!------------------------------------------------------------------------------
       CovMetric = CtrMetric
       CALL InvertMatrix( CovMetric,3 )

       DO i=1,3
         DO j=1,3
            s = 0.0d0
            t = 0.0d0

            DO k=1,3
               s = s + CovMetric(i,k) * dVelodx(k,j) + &
                       CovMetric(j,k) * dVelodx(k,i)

              t = t + CtrMetric(j,k) * dVelodx(i,k) + &
                      CtrMetric(i,k) * dVelodx(j,k)

              DO l=1,3
                s = s - CovMetric(i,k) * Symb(l,j,k) * Velo(l)
                s = s - CovMetric(j,k) * Symb(l,i,k) * Velo(l)

                t = t - CtrMetric(j,k) * Symb(l,k,i) * Velo(l)
                t = t - CtrMetric(i,k) * Symb(l,k,j) * Velo(l)
              END DO
           END DO
           SecInv = SecInv + s * t
         END DO
       END DO
!------------------------------------------------------------------------------

     END IF
!------------------------------------------------------------------------------
   END FUNCTION SecondInvariant
!------------------------------------------------------------------------------

!------------------------------------------------------------------------------
!> Rotate a (6x6) tensor based on a trasformation matrix
!> Note: Copied from Stress.F90 
!------------------------------------------------------------------------------
!------------------------------------------------------------------------------
  SUBROUTINE RotateMatrix3D(C,T)
!------------------------------------------------------------------------------
    IMPLICIT NONE

    REAL(KIND=dp) :: T(:,:), C(:,:), CT(3,3,3,3)
    INTEGER :: i,j,p,q,r,s
    INTEGER :: I1(6) = [ 1,2,3,1,2,1 ], I2(6) = [ 1,2,3,2,3,3 ]

    !
    ! Convert C-matrix to 4 index elasticity tensor:
    ! ----------------------------------------------
    CT = 0.0d0
    DO i=1,6
      p = I1(i)
      q = I2(i)
      DO j=1,6
        r = I1(j)
        s = I2(j)
        CT(p,q,r,s) = C(i,j)
        CT(p,q,s,r) = C(i,j)
        CT(q,p,r,s) = C(i,j)
        CT(q,p,s,r) = C(i,j)
      END DO
    END DO

    !
    ! Rotate the tensor:
    ! ------------------
    CALL Rotate4IndexTensor( CT, T, 3 )

    !
    ! Convert back to matrix form:
    ! ----------------------------
    DO i=1,6
      p = I1(i)
      q = I2(i)
      DO j=1,6
        r = I1(j)
        s = I2(j)
        C(i,j) = CT(p,q,r,s)
      END DO
    END DO
!------------------------------------------------------------------------------
  END SUBROUTINE RotateMatrix3D
!------------------------------------------------------------------------------

!------------------------------------------------------------------------------
!> Rotate a (3x3x3x3) tensor based on a trasformation matrix
!> Note: Copied from Stress.F90
!------------------------------------------------------------------------------
!------------------------------------------------------------------------------
   SUBROUTINE Rotate4IndexTensor( C, T, dim )
!------------------------------------------------------------------------------
     INTEGER :: dim
     REAL(KIND=dp) :: C(:,:,:,:),T(:,:)
!------------------------------------------------------------------------------
     INTEGER :: i,j
     REAL(KIND=dp) :: C1(dim,dim,dim,dim)
!------------------------------------------------------------------------------
     C1 = 0
     DO i=1,dim
       DO j=1,dim
         C1(:,:,:,i) = C1(:,:,:,i) + T(i,j)*C(:,:,:,j)
       END DO
     END DO

     C = 0
     DO i=1,dim
       DO j=1,dim
         C(:,:,i,:) = C(:,:,i,:) + T(i,j)*C1(:,:,j,:)
       END DO
     END DO

     C1 = 0
     DO i=1,dim
       DO j=1,dim
         C1(:,i,:,:) = C1(:,i,:,:) + T(i,j)*C(:,j,:,:)
       END DO
     END DO

     C = 0
     DO i=1,dim
       DO j=1,dim
         C(i,:,:,:) = C(i,:,:,:) + T(i,j)*C1(j,:,:,:)
       END DO
     END DO
!------------------------------------------------------------------------------
   END SUBROUTINE Rotate4IndexTensor
!------------------------------------------------------------------------------


!------------------------------------------------------------------------------
!> Returns effective anisotropic viscosity for Navier-Stokes equation.
!> Note: Based on the EffectiveViscosity function in MaterialModels.f90
!------------------------------------------------------------------------------
   FUNCTION AnisotropicEffectiveViscosity( Viscosity,Density,Ux,Uy,Uz,Element, &
        Nodes,n,nd,u,v,w, muder, LocalIP ) RESULT(mu_tensor)
     !------------------------------------------------------------------------------

     USE ModelDescription

     REAL(KIND=dp)  :: Viscosity,Density,u,v,w,mu,Ux(:),Uy(:),Uz(:)
     REAL(KIND=dp), OPTIONAL :: muder
     TYPE(Nodes_t)  :: Nodes
     INTEGER :: n,nd
     INTEGER, OPTIONAL :: LocalIP
     TYPE(Element_t),POINTER :: Element

     !------------------------------------------------------------------------------
     REAL(KIND=dp) :: Basis(nd),dBasisdx(nd,3)
     REAL(KIND=dp) :: ss,s,SqrtMetric,SqrtElementMetric,Velo(3)
     REAL(KIND=dp) :: Metric(3,3), dVelodx(3,3), CtrMetric(3,3), &
          Symb(3,3,3), dSymb(3,3,3,3) 

     INTEGER :: i,j,k
     LOGICAL :: stat,GotIt,UseEUsrf=.FALSE.

     CHARACTER(LEN=MAX_NAME_LEN) :: ViscosityFlag, TemperatureName, EnhcmntFactFlag
     TYPE(ValueList_t), POINTER :: Material
     REAL(KIND=dp) :: x, y, z, c1n(n), c2n(n), c3n(n), c4n(n), &
          c1, c2, c3, c4, UnitNorm, TransformMatrix(3,3), mu_tensor(6,6) 

     REAL(KIND=dp), POINTER :: Uwrk(:,:)


     !------------------------------------------------------------------------------
     mu = Viscosity

     k = ListGetInteger( CurrentModel % Bodies(Element % BodyId) % Values, 'Material', &
          minv=1, maxv=CurrentModel % NumberOFMaterials )

     Material => CurrentModel % Materials(k) % Values

     ViscosityFlag = ListGetString( Material,'Viscosity Model', GotIt)

     IF(.NOT. gotIt) RETURN
     !------------------------------------------------------------------------------
     !    Basis function values & derivatives at the calculation point
     !------------------------------------------------------------------------------
     stat = ElementInfo( Element,Nodes,u,v,w, &
          SqrtElementMetric, Basis,dBasisdx )
     !------------------------------------------------------------------------------
     !   Coordinate system dependent information
     !------------------------------------------------------------------------------
     x = SUM( Nodes % x(1:n) * Basis(1:n) )
     y = SUM( Nodes % y(1:n) * Basis(1:n) )
     z = SUM( Nodes % z(1:n) * Basis(1:n) )
     CALL CoordinateSystemInfo( Metric,SqrtMetric,Symb,dSymb,x,y,z )
     !------------------------------------------------------------------------------
     DO j=1,3
        dVelodx(1,j) = SUM( Ux(1:nd)*dBasisdx(1:nd,j) )
        dVelodx(2,j) = SUM( Uy(1:nd)*dBasisdx(1:nd,j) )
        dVelodx(3,j) = SUM( Uz(1:nd)*dBasisdx(1:nd,j) )
     END DO

     Velo(1) = SUM( Basis(1:nd) * Ux(1:nd) )
     Velo(2) = SUM( Basis(1:nd) * Uy(1:nd) )
     Velo(3) = SUM( Basis(1:nd) * Uz(1:nd) )

     ! This is the square of shearrate which results to 1/2 in exponent
     ! Also the derivative is taken with respect to the square
     !-------------------------------------------------------------------
     ss = 0.5_dp * SecondInvariant(Velo,dVelodx,Metric,Symb)

     SELECT CASE( ViscosityFlag )

     CASE('orthotropic non linear')
        ! Read Fluidity Prefactor
        c1n = ListGetReal( Material, 'Fluidity', n, Element % NodeIndexes )
        c1 = SUM( Basis(1:n) * c1n(1:n) )

        ! Read Power Law Exponent
        c2n = ListGetReal( Material, 'Power law Exponent', n, Element % NodeIndexes )
        c2 = SUM( Basis(1:n) * c2n(1:n) )
        
        ! Read Critical Shear Rate and set the limit if needed
        c3n = ListGetReal( Material, 'Critical Shear Rate',n, Element % NodeIndexes,gotIt )
        IF (GotIt) THEN
           c3 = SUM( Basis(1:n) * c3n(1:n) )
           IF(ss < c3**2) THEN
              ss = c3**2
           END IF
        END IF

        ! Read Anisotropy parameter
        c4n = ListGetReal( Material, 'Anisotropy Parameter', n, Element % NodeIndexes )
        c4 = SUM( Basis(1:n) * c4n(1:n) )


        ! Read Transform Matrix based on the Material Coordinates Unit Vectors
        DO i=1,3
          IF( i == 1 ) THEN
            CALL GetConstRealArray( Material, UWrk, &
                'Material Coordinates Unit Vector 1', stat, Element )
          ELSE IF( i == 2 ) THEN
            CALL GetConstRealArray( Material, UWrk, &
                'Material Coordinates Unit Vector 2', stat, Element )
          ELSE                
            CALL GetConstRealArray( Material, UWrk, &
                'Material Coordinates Unit Vector 3', stat, Element )
          END IF
          IF( stat ) THEN
            UnitNorm = SQRT( SUM( Uwrk(1:3,1)**2 ) )
            IF( UnitNorm < EPSILON( UnitNorm ) ) THEN
              CALL Fatal('AnisotropicEffectiveViscosity','Given Material Coordinate Unit Vector too short!')
            END IF
            TransformMatrix(i,1:3) = Uwrk(1:3,1) / UnitNorm
          END IF
        END DO

        ! Compute equivalent Viscosity prefactor
        mu = (1.0_dp/c1)**(1.0_dp/c2)

        ! Create Viscosity tensor in the grain frame and with Voigt Notation
        ! Taken from Gagliardini and Meyissonnier 1999 
        mu_tensor = 0.0_dp
        mu_tensor(1,1) = mu * c4
        mu_tensor(2,2) = mu * c4
        mu_tensor(3,3) = mu * c4
        mu_tensor(4,4) = mu * c4  ! Not a "real" Voigt Notation as there is no factor two on the stresses
        mu_tensor(5,5) = mu       ! Not a "real" Voigt Notation as there is no factor two on the stresses 
        mu_tensor(6,6) = mu       ! Not a "real" Voigt Notation as there is no factor two on the stresses

        ! Rotate to general frame
        CALL RotateMatrix3D(mu_tensor,TransformMatrix)

        ! Multiply by strain rate invariant for non-linearity
        mu_tensor = mu_tensor * ss**(((1.0_dp/c2) -1.0_dp)/2.0_dp)

     CASE DEFAULT
        CALL WARN('AnisotropicEffectiveViscosity','Unknown material model')

     END SELECT

!------------------------------------------------------------------------------
   END FUNCTION AnisotropicEffectiveViscosity
!------------------------------------------------------------------------------
END MODULE AnisotropicMaterialModels
